VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "DriveInfo"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'=========
'DriveInfo
'=========
'
'PUBLIC METHODS, PROPERTIES, EVENTS, ETC.
'----------------------------------------
'
'Set the property Letter As String to a drive letter to look up
'the drive's attributes.
'
'Interrogate the other properties as desired:
'
'   BusType
'   BusTypeText
'   DeviceType
'   DeviceTypeText
'   Letter
'   ProductId
'   ProductRevision
'   Removable
'   SerialNumber
'   VendorId
'
'
'SUPPORTED VERSIONS OF WINDOWS
'-----------------------------
'
'Uses API calls that are not supported until Windows 98.  Some calls
'may not work at all on Win9x systems.
'
'Not fully tested on Windows XP.  Not tested on Windows 2000 or
'earlier.
'
'Windows Vista will report SATA drives as ATA.  Windows 7 and later
'will report SATA drives in PATA mode as ATA, in SATA mode as SATA.
'
'
'VERSION 2
'---------
'
'Beginning with Windows 8, Microsoft has corrected drive serial number
'byte order.  Dertect the OS version to decide whether or not to
'apply the corerction here.
'
'Tested on Windows 10.
'
'
'VERSION 2.1
'-----------
'
'Testing USB HDD caused an "Out of memory" error 7.  Now checking for
'Size > 0 on several fetched attribute values before attempting to
'use them.
'
'Testing a USB HDD retrieved "garbage" serial number value.  Commented
'out old logic to treat some values as a string of hex digits (find
'comment below that starts with '##).
'
'Testing a USB HDD on Win7 was reported to swap SerialNumber characters.
'Now swap characters when (BeforeWin8 And mBusType <> BusTypeUsb).
'
'
'VERSION 2.2
'-----------
'
'Add back hex-digit collapsing but (1.) do it after the conditional
'byte swapping, and (2.) test the result for a plausible serial number
'(alphanumeric, -, and space chars) and if not plausible keep the "hex
'digits" as the serial number.
'
'
'VERSION 2.3
'-----------
'
'Move the conditional byte swapping to AFTER detecting hex digits.
'

Private Const GENERIC_READ = &H80000000
Private Const GENERIC_WRITE = &H40000000
Private Const FILE_SHARE_READ = &H1
Private Const FILE_SHARE_WRITE = &H2
Private Const OPEN_EXISTING = 3
Private Const INVALID_HANDLE_VALUE = -1

Private Const IOCTL_STORAGE_QUERY_PROPERTY = &H2D1400

Private Type STORAGE_PROPERTY_QUERY
    PropertyId As Integer
    QueryType As Integer
    AdditionalParameters(7) As Byte
End Type

Public Enum SCSI_DEVICE_TYPES        'SCSI device types:
    DeviceTypeDisk = &H0             'Direct-access device (e.g., magnetic disk)
    DeviceTypeTape = &H1             'Sequential-access device (e.g., magnetic tape)
    DeviceTypePrinter = &H2          'Printer device
    DeviceTypeProcessor = &H3        'Processor device
    DeviceTypeWriteOnce = &H4        'Write-once device
    DeviceTypeCDROM = &H5            'CD/DVD-ROM device
    DeviceTypeScanner = &H6          'Scanner device
    DeviceTypeOpticalStore = &H7     'Optical memory device (e.g., some optical disks)
    DeviceTypeChanger = &H8          'Medium Changer (e.g. jukeboxes)
    DeviceTypeComm = &H9             'Communications device
    DeviceTypeASC_IT8_1 = &HA        'Defined by ASC IT8 (Graphic arts pre-press devices)
    DeviceTypeASC_IT8_2 = &HB        'Defined by ASC IT8 (Graphic arts pre-press devices)
    DeviceTypeStoreArray = &HC       'Storage array controller device (e.g., RAID)
    DeviceTypeEnclosure = &HD        'Enclosure services device
    DeviceTypeSimpleDirect = &HE     'Simplified direct-access device (e.g., magnetic disk)
    DeviceTypeOpticalCard = &HF      'Optical card reader/writer device
    DeviceTypeBridgeExpander = &H10  'Reserved for bridging expanders
    DeviceTypeObjectStore = &H11     'Object-based Storage Device
    DeviceTypeAutomation = &H12      'Automation/Drive Interface
    '&H13-&H1D - reserved
    DeviceTypeWellKnownLU = &H1E     'Well known logical unit
    DeviceTypeUnknown = &H1F         'Unknown or no device type
End Enum
#If False Then
Dim DeviceTypeDisk, DeviceTypeTape, DeviceTypePrinter, DeviceTypeProcessor
Dim DeviceTypeWriteOnce, DeviceTypeCDROM, DeviceTypeScanner
Dim DeviceTypeOpticalStore, DeviceTypeChanger, DeviceTypeComm, DeviceTypeASC_IT8_1
Dim DeviceTypeASC_IT8_2, DeviceTypeStoreArray, DeviceTypeEnclosure
Dim DeviceTypeSimpleDirect, DeviceTypeOpticalCard, DeviceTypeBridgeExpander
Dim DeviceTypeObjectStore, DeviceTypeAutomation, DeviceTypeWellKnownLU
Dim DeviceTypeUnknown
#End If

Private Const DEVICE_TYPES_TEXT = _
        "Disk|Tape|Printer|Processor|WriteOnce|CDROM|Scanner" _
      & "|OpticalStore|Changer|Comm|ASC IT8 1|ASC IT8 2" _
      & "|StoreArray|Enclosure|SimpleDirect|OpticalCard" _
      & "|BridgeExpander|ObjectStore|Automation|0x13|0x14" _
      & "|0x15|0x16|0x17|0x18|0x19|0x1A|0x1B|0x1C|0x1D" _
      & "|WellKnownLU|Unknown"

Public Enum BUS_TYPES
    BusTypeUnknown = &H0           'Indicates an unknown bus type.
    BusTypeScsi = &H1              'Indicates a small computer system interface (SCSI) bus.
    BusTypeAtapi = &H2             'Indicates an AT Attachment Packet Interface (ATAPI) bus:
                                   '    Optical drive
    BusTypeAta = &H3               'Indicates an advanced technology attachment (ATA) bus:
                                   '    PATA hard disk drive
                                   '    SATA hard disk drive (controller in PATA mode)
                                   '    SATA hard disk drive (controller in SATA mode) - prior to Win7.
    BusType1394 = &H4              'Indicates an IEEE 1394 bus.
    BusTypeSsa = &H5               'Indicates a serial storage architecture (SSA) bus.
    BusTypeFibre = &H6             'Indicates a fibre channel bus type.
    BusTypeUsb = &H7               'Indicates a USB bus type.
    BusTypeRAID = &H8              'Indicates a bus for a redundant array of independent disks (RAID).
    BusTypeiScsi = &H9             'Indicates an iSCSI bus.
    BusTypeSas = &HA               'Indicates a serial-attached SCSI bus.
    BusTypeSata = &HB              'Indicates a serial ATA bus:
                                   '    SATA hard disk drive (controller in SATA mode) - starting in Win7.
    BusTypeSd = &HC                'Indicates a secure digital bus.
    BusTypeMmc = &HD               'Indicates a multimedia card bus.
    BusTypeVirtual = &HE           'Indicates a virtual bus type.
    BusTypeFileBackedVirtual = &HF 'Indicates a file-backed virtual bus type.
    BusTypeMax = &H10              'Indicates the maximum value for this value.
    BusTypeMaxReserved = &H7F      'The maximum value of the storage bus type range.
End Enum
#If False Then
Dim BusTypeUnknown, BusTypeScsi, BusTypeAtapi, BusTypeAta, BusType1394, BusTypeSsa
Dim BusTypeFibre, BusTypeUsb, BusTypeRAID, BusTypeiScsi, BusTypeSas, BusTypeSata
Dim BusTypeSd, BusTypeMmc, BusTypeVirtual, BusTypeFileBackedVirtual, BusTypeMax
Dim BusTypeMaxReserved
#End If

Private Const BUS_TYPES_TEXT = _
        "Unknown|SCSI|ATAPI|ATA|IEEE 1394|SSA|Fibre" _
      & "|USB|RAID|iSCSI|saSCSI|SATA|SD|MMC|Virtual" _
      & "|FileBackedVirtual"

Private Type STORAGE_DEVICE_DESCRIPTOR
    Version As Long '0
    Size As Long '4
    DeviceType As Byte '8
    DeviceTypeModifier As Byte '9
    RemovableMedia As Byte '10
    CommandQueueing As Byte '11
    VendorIdOffset As Long '12
    ProductIdOffset As Long '16
    ProductRevisionOffset As Long '20
    SerialNumberOffset As Long '24
    BusType As BUS_TYPES '28
    RawPropertiesLength As Long '32
    RawDeviceProperties(36 To 255) As Byte '36
End Type

Private Declare Function CloseHandle Lib "kernel32" (ByVal hObject As Long) As Long

Private Declare Function CreateFile Lib "kernel32" Alias "CreateFileA" ( _
    ByVal lpFileName As String, _
    ByVal dwDesiredAccess As Long, _
    ByVal dwShareMode As Long, _
    ByVal lpSecurityAttributes As Long, _
    ByVal dwCreationDisposition As Long, _
    ByVal dwFlagsAndAttributes As Long, _
    ByVal hTemplateFile As Long) As Long

Private Declare Function DeviceIoControl Lib "kernel32" ( _
    ByVal hDevice As Long, _
    ByVal dwIoControlCode As Long, _
    ByVal lpInBuffer As Long, _
    ByVal nInBufferSize As Long, _
    ByVal lpOutBuffer As Long, _
    ByVal nOutBufferSize As Long, _
    ByRef lpBytesReturned As Long, _
    ByVal lpOverlapped As Long) As Long

Private Declare Function GetVersion Lib "kernel32" () As Long

Private Declare Function lstrlen Lib "kernel32" Alias "lstrlenA" ( _
    ByVal lpString As Long) As Long

Private UsbFlashSerial As UsbFlashSerial
Private BeforeWin8 As Boolean

Private mBusType As BUS_TYPES
Private mDeviceType As SCSI_DEVICE_TYPES
Private mLetter As String
Private mProductId As String
Private mProductRevision As String
Private mRemovable As Boolean
Private mSerialNumber As String
Private mVendorId As String

Public Property Get BusType() As BUS_TYPES
    BusType = mBusType
End Property

Public Property Get BusTypeText() As String
    If mBusType >= BusTypeMax Then
        BusTypeText = "?"
    Else
        BusTypeText = Split(BUS_TYPES_TEXT, "|")(mBusType)
    End If
End Property

Public Property Get DeviceType() As SCSI_DEVICE_TYPES
    DeviceType = mDeviceType
End Property

Public Property Get DeviceTypeText() As String
    If mDeviceType > DeviceTypeUnknown Then
        DeviceTypeText = "?"
    Else
        DeviceTypeText = Split(DEVICE_TYPES_TEXT, "|")(mDeviceType)
    End If
End Property

Public Property Get Letter() As String
    Letter = mLetter
End Property

Public Property Let Letter(ByVal RHS As String)
    Dim hDevice As Long
    Dim spQuery As STORAGE_PROPERTY_QUERY
    Dim sdDesc As STORAGE_DEVICE_DESCRIPTOR
    Dim BytesReturned As Long
    Dim Size As Long
    Dim Bytes() As Byte
    Dim I As Integer
    Dim Temp As Byte
    Dim Temps() As Byte
    Dim UsbSerial As String
    
    RHS = UCase$(Left$(RHS, 1))
    If RHS Like "[A-Z]" Then
        mLetter = RHS
        
        hDevice = CreateFile("\\.\" & mLetter & ":", _
                             0, _
                             FILE_SHARE_READ Or FILE_SHARE_WRITE, _
                             0, _
                             OPEN_EXISTING, _
                             0, _
                             0)
        If hDevice = INVALID_HANDLE_VALUE Then
            Err.Raise &H80047100, TypeName(Me), "CreateFile error " & CStr(Err.LastDllError)
        Else
            If DeviceIoControl(hDevice, _
                               IOCTL_STORAGE_QUERY_PROPERTY, _
                               VarPtr(spQuery), _
                               Len(spQuery), _
                               VarPtr(sdDesc), _
                               Len(sdDesc), _
                               BytesReturned, _
                               0) = 0 Then
                Err.Raise &H80047104, _
                          TypeName(Me), _
                          "DeviceIoControl error " & CStr(Err.LastDllError)
            Else
                CloseHandle hDevice
                With sdDesc
                    mBusType = .BusType
                    mDeviceType = .DeviceType
                    mRemovable = .RemovableMedia <> 0
                    
                    If .ProductIdOffset > 0 Then
                        Size = lstrlen(VarPtr(.RawDeviceProperties(.ProductIdOffset)))
                        If Size > 0 Then
                            Bytes = MidB$(.RawDeviceProperties, _
                                          .ProductIdOffset - LBound(.RawDeviceProperties) + 1, _
                                          Size)
                            mProductId = Trim$(StrConv(Bytes, vbUnicode))
                        Else
                            mProductId = vbNullString
                        End If
                    Else
                        mProductId = vbNullString
                    End If
                    
                    If .ProductRevisionOffset > 0 Then
                        Size = lstrlen(VarPtr(.RawDeviceProperties(.ProductRevisionOffset)))
                        If Size > 0 Then
                            Bytes = MidB$(.RawDeviceProperties, _
                                          .ProductRevisionOffset - LBound(.RawDeviceProperties) + 1, _
                                          Size)
                            mProductRevision = Trim$(StrConv(Bytes, vbUnicode))
                        Else
                            mProductRevision = vbNullString
                        End If
                    Else
                        mProductRevision = vbNullString
                    End If
                    
                    If .SerialNumberOffset > 0 Then
                        Size = lstrlen(VarPtr(.RawDeviceProperties(.SerialNumberOffset)))
                        If Size > 0 Then
                            If .RawDeviceProperties(.SerialNumberOffset) >= 32 Then
                                Bytes = MidB$(.RawDeviceProperties, _
                                              .SerialNumberOffset - LBound(.RawDeviceProperties) + 1, _
                                              Size)
                                If Size Mod 2 = 0 _
                                    And Not (StrConv(Bytes, vbUnicode) Like "*[!0-9A-Fa-f]*") Then
                                    'Serial number appears to be given in hex digits, collapse Hex
                                    'digits to characters:
                                    ReDim Temps(Size \ 2 - 1)
                                    For I = 0 To Size - 2 Step 2
                                        Temps(I \ 2) = AscHex2Byte(Bytes(I)) * CByte(&H10) _
                                                     + AscHex2Byte(Bytes(I + 1))
                                    Next
                                    If Not StrConv(Temps, vbUnicode) Like "*[! -0-9A-Za-z]*" Then
                                        'Only alphanumeric, - and space so good after collapsing:
                                        Bytes = Temps
                                        Size = Size \ 2
                                    End If
                                End If
                                If BeforeWin8 And mBusType <> BusTypeUsb Then
                                    'Swap character order:
                                    For I = 0 To Size - 2 Step 2
                                        Temp = Bytes(I)
                                        Bytes(I) = Bytes(I + 1)
                                        Bytes(I + 1) = Temp
                                    Next
                                End If
                                mSerialNumber = Trim$(StrConv(LeftB$(Bytes, Size), vbUnicode))
                            ElseIf .RawDeviceProperties(.SerialNumberOffset) = 0 Then
                                mSerialNumber = vbNullString
                            Else
                                mSerialNumber = "binary?"
                            End If
                        Else
                            mSerialNumber = vbNullString
                        End If
                    Else
                        mSerialNumber = vbNullString
                    End If
                    
                    If .VendorIdOffset > 0 Then
                        Size = lstrlen(VarPtr(.RawDeviceProperties(.VendorIdOffset)))
                        If Size > 0 Then
                            Bytes = MidB$(.RawDeviceProperties, _
                                          .VendorIdOffset - LBound(.RawDeviceProperties) + 1, _
                                          Size)
                            mVendorId = Trim$(StrConv(Bytes, vbUnicode))
                        Else
                            mVendorId = vbNullString
                        End If
                    Else
                        mVendorId = vbNullString
                    End If
                End With
                
                If mRemovable And mBusType = BusTypeUsb And mDeviceType = DeviceTypeDisk Then
                    'Special case, USB flash drives?  Try using UsbFlashSerial:
                    UsbSerial = UsbFlashSerial.Lookup(mLetter)
                    If Len(UsbSerial) Then mSerialNumber = UsbSerial
                End If
            End If
        End If
    Else
        Err.Raise &H80047108, TypeName(Me), "Bad value for Letter"
    End If
End Property

Public Property Get ProductId() As String
    ProductId = mProductId
End Property

Public Property Get ProductRevision() As String
    ProductRevision = mProductRevision
End Property

Public Property Get Removable() As Boolean
    Removable = mRemovable
End Property

Public Property Get SerialNumber() As String
    SerialNumber = mSerialNumber
End Property

Public Property Get VendorId() As String
    VendorId = mVendorId
End Property

Private Function AscHex2Byte(ByVal AscHex As Byte) As Byte
    'Convert a Byte containing a single ASCII Hex digit to the corresponding Byte value.
    'Assume valid values ("0"-"9", "A"-"Z", "a"-"z").
    Const Asc0 As Byte = &H30
    Const AscUpperA As Byte = &H41
    Const AscLowera As Byte = &H61
    
    If AscHex < AscUpperA Then
        AscHex2Byte = CByte(AscHex - Asc0)
    ElseIf AscHex < AscLowera Then
        AscHex2Byte = CByte(AscHex - AscUpperA + 10)
    Else
        AscHex2Byte = CByte(AscHex - AscLowera + 10)
    End If
End Function

Private Sub Class_Initialize()
    Const WIN_8 As Long = 6002
    Dim OS As Long
    
    OS = GetVersion()
    BeforeWin8 = (OS And &HFF&) * 1000 + (OS And &HFF00&) \ &H100& < WIN_8
    Set UsbFlashSerial = New UsbFlashSerial
End Sub
